package com.tencent.subscribe.main.datastream;

import com.google.protobuf.ByteString;
import com.tencent.subscribe.beans.RecordMsgObject;
import com.tencent.subscribe.pb.SubscribeProtobufData;
import com.tencent.subscribe.serialize.DeserializeProtobufToRecordMsgObject;
import org.apache.commons.cli.*;
import org.apache.flink.api.common.state.ValueState;
import org.apache.flink.api.common.state.ValueStateDescriptor;
import org.apache.flink.api.common.typeinfo.Types;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.CheckpointingMode;
import org.apache.flink.streaming.api.datastream.DataStream;
import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
import org.apache.flink.streaming.api.functions.KeyedProcessFunction;
import org.apache.flink.streaming.connectors.kafka.FlinkKafkaConsumer;
import org.apache.flink.streaming.connectors.kafka.FlinkKafkaConsumerBase;
import org.apache.flink.util.Collector;
import org.apache.kafka.clients.CommonClientConfigs;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.common.config.SaslConfigs;
import org.apache.kafka.common.header.Header;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import javax.xml.bind.DatatypeConverter;
import java.io.ByteArrayOutputStream;
import java.time.Instant;
import java.time.LocalDateTime;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import java.util.TimeZone;

public class SubscribeFromKafkaFlinkForDataStream {

    private static final Logger log = LoggerFactory.getLogger(SubscribeFromKafkaFlinkForDataStream.class);

    public static void main(final String[] args) {
        Option option_brokers = Option.builder()
                .required(true)
                .desc("The brokers option")
                .longOpt("brokers")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_topic = Option.builder()
                .required(true)
                .desc("The topic option")
                .longOpt("topic")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_group = Option.builder()
                .required(true)
                .desc("The group option")
                .longOpt("group")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_user = Option.builder()
                .required(false)
                .desc("The user option")
                .longOpt("user")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_password = Option.builder()
                .required(false)
                .desc("The password option")
                .longOpt("password")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_trans = Option.builder()
                .required(false)
                .desc("The trans2sql option")
                .longOpt("trans2sql")
                .build();

        Options options = new Options()
                .addOption(option_brokers)
                .addOption(option_topic)
                .addOption(option_user)
                .addOption(option_password)
                .addOption(option_group)
                .addOption(option_trans);

        CommandLineParser parser = new DefaultParser();

        try {
            CommandLine commandLine = parser.parse(options, args);

            run(
                    commandLine.getOptionValue("brokers"),
                    commandLine.getOptionValue("topic"),
                    commandLine.getOptionValue("group"),
                    commandLine.getOptionValue("user"),
                    commandLine.getOptionValue("password"),
                    commandLine.hasOption("trans2sql")
            );

        } catch (Exception exception) {
            System.out.print("error: ");
            System.out.println(exception);
        }
    }

    private static void run(
            String brokers,
            String topic,
            String group,
            String user,
            String password,
            boolean trans2sql) throws Exception {
        Properties props = new Properties();
        props.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, brokers);
        props.put(ConsumerConfig.GROUP_ID_CONFIG, group);
        props.put(ConsumerConfig.CLIENT_ID_CONFIG, group + "_" + Instant.now().getEpochSecond());
        props.put(ConsumerConfig.ENABLE_AUTO_COMMIT_CONFIG, true);
        props.put(ConsumerConfig.AUTO_COMMIT_INTERVAL_MS_CONFIG, 5000);
        props.put(ConsumerConfig.ISOLATION_LEVEL_CONFIG, "read_committed");
        props.put(ConsumerConfig.AUTO_OFFSET_RESET_CONFIG, "earliest");
        props.put(ConsumerConfig.FETCH_MAX_BYTES_CONFIG, 20 << 20); // kafka消息上限20M
        props.setProperty(FlinkKafkaConsumerBase.KEY_PARTITION_DISCOVERY_INTERVAL_MILLIS, "10");

        if (null != user && !user.equals("")) {
            props.put(CommonClientConfigs.SECURITY_PROTOCOL_CONFIG, "SASL_PLAINTEXT");
            props.put(SaslConfigs.SASL_MECHANISM, "SCRAM-SHA-512");
            props.put(SaslConfigs.SASL_JAAS_CONFIG,
                    "org.apache.kafka.common.security.scram.ScramLoginModule required " +
                            "  username=\"" + user + "\"" +
                            "  password=\"" + password + "\";");
        }

        System.out.println("flink kafka consumer is ready to consume...");
        System.out.println("topic :" + topic);
        System.out.println("group :" + group);
        System.out.println("user :" + user);
        System.out.println("password :" + password);
        System.out.println("brokers :" + brokers);
        System.out.println("trans2sql :" + trans2sql);

        if (null != user && !user.equals("")) {
            props.put(CommonClientConfigs.SECURITY_PROTOCOL_CONFIG, "SASL_PLAINTEXT");
            props.put(SaslConfigs.SASL_MECHANISM, "SCRAM-SHA-512");
            props.put(SaslConfigs.SASL_JAAS_CONFIG,
                    "org.apache.kafka.common.security.scram.ScramLoginModule required " +
                            "  username=\"" + user + "\"" +
                            "  password=\"" + password + "\";");
        }

        StreamExecutionEnvironment env = StreamExecutionEnvironment.getExecutionEnvironment();
        FlinkKafkaConsumer<RecordMsgObject> consumer = new FlinkKafkaConsumer<>(topic, new DeserializeProtobufToRecordMsgObject(), props);
//        consumer.setStartFromGroupOffsets();
        // 从最新位点开始消费数据
        consumer.setStartFromLatest();
        consumer.setCommitOffsetsOnCheckpoints(true);

        DataStream<RecordMsgObject> stream = env.addSource(consumer);

        //将收到的消息按照parttition分区处理
        stream.keyBy(RecordMsgObject::getPartition)
                .process(new SubscribeMsgProcess(trans2sql)).setParallelism(1);

        env.enableCheckpointing(5000, CheckpointingMode.AT_LEAST_ONCE);

        // zzz
        env.execute("dts subscribe flink consumer");
    }

    // 按照分区消费，每个分区拿到一个envlope，我们需要拆解这个envlope，注意当前类中的所有fields必须要保证全部实现Serializable接口
    public static class SubscribeMsgProcess extends KeyedProcessFunction<String, RecordMsgObject, Object> {
        HashMap<String, ByteArrayOutputStream> shardMsgMap = new HashMap<>();
        HashMap<String, Long> shardLastSeqMap = new HashMap<>();
        private ValueState<Long> maxSeqState;
        private final boolean trans2sql;
        private final static String partitionSeqKey = "ps";
        private final static String ShardIdKey = "ShardId";

        public SubscribeMsgProcess(boolean trans2sql) {
            this.trans2sql = trans2sql;
        }

        @Override
        public void open(Configuration parameters) throws Exception {
            maxSeqState = getRuntimeContext().getState(new ValueStateDescriptor<Long>("maxSeqState", Types.LONG, (long) 0));
        }

        @Override
        public void processElement(RecordMsgObject record, Context context, Collector<Object> collector) throws Exception {

            // 获取每个消息的分区排序
            long ps = getPartitionSeq(record);

            // 拿到每个消息的shardId
            String shardId = getShardId(record);
            System.out.println("shardId:" + shardId);
            Long shardLastSeq = shardLastSeqMap.get(shardId);
            if (shardLastSeq == null || ps == 1) {
                shardLastSeqMap.put(shardId, (long) 1);
            } else if (ps <= shardLastSeq) {
                System.out.println("duplicated message: " + record);
                return;
            } else if (ps != shardLastSeq + 1 && shardLastSeq != 0) {
                throw new IllegalStateException("the partition seq discontinuous, last: " + shardLastSeq + ", current: " + ps);
            }
            shardLastSeqMap.put(shardId, ps);

            // 拿到envlope，判断是不是还有消息
            SubscribeProtobufData.Envelope envelope = SubscribeProtobufData.Envelope.parseFrom(record.value);
            if (1 != envelope.getVersion()) {
                throw new IllegalStateException(String.format("unsupported version: %d", envelope.getVersion()));
            }


            // 现在默认就是一个分片，有可能会有多个分片，这里就需要单独处理了
            ByteString envelopeData = envelope.getData();
            // 拿到Entries
            SubscribeProtobufData.Entries entries;
            if (1 == envelope.getTotal()) {
                entries = SubscribeProtobufData.Entries.parseFrom(envelopeData.toByteArray());
            } else {
                entries = SubscribeProtobufData.Entries.parseFrom(shardMsgMap.get(shardId).toByteArray());
                // Pay attention to memory recycling for shardMsgMap
                shardMsgMap.remove(shardId);
            }
            // 遍历每个Entry,根据Entry类型去处理
            for (SubscribeProtobufData.Entry entry : entries.getItemsList()) {
                onEntry(record.partition, record.offset, ps, entry, trans2sql);
            }
            maxSeqState.update(ps);
        }

        private static String getShardId(RecordMsgObject msg) {
            Iterator<Header> it = msg.headers.headers(ShardIdKey).iterator();
            if (it.hasNext()) {
                Header h = it.next();
                return new String(h.value());
            }
            throw new IllegalStateException(ShardIdKey + " does not exists");
        }

        private static long getPartitionSeq(RecordMsgObject msg) {
            Iterator<Header> it = msg.headers.headers(partitionSeqKey).iterator();
            if (it.hasNext()) {
                Header h = it.next();
                return Long.parseUnsignedLong(new String(h.value()));
            }
            throw new IllegalStateException(partitionSeqKey + " does not exists");
        }


    }

    private static void onEntry(int partitionNum, long offset, long partitionSeq, SubscribeProtobufData.Entry record, boolean trans2sql) throws Exception {
        log.info("partition:" + partitionNum + ", offset:" + offset + ", partitionSeq:" + partitionSeq
                + ", " + record.toString());
        if (trans2sql) {
            trans2SQL(partitionNum, offset, partitionSeq, record);
        }
    }

    private static void trans2SQL(int partitionNum, long offset, long partitionSeq, SubscribeProtobufData.Entry record) throws Exception {
        System.out.println(">>>>>>>>>>>>>>>>>>>>trans2SQL>>>>>>>>>>>>>>>>>>>>>");
        SubscribeProtobufData.Header header = record.getHeader();
        System.out.println(
                String.format("-->[partition: %d, offset: %d, partitionSeq: %d] [%s:%d], happenedAt: %s",
                        partitionNum,
                        offset,
                        partitionSeq,
                        header.getFileName(), header.getPosition(),
                        LocalDateTime.ofInstant(Instant.ofEpochSecond(header.getTimestamp()),
                                TimeZone.getDefault().toZoneId())));
        switch (record.getHeader().getMessageType()) {
            case HEARTBEAT:
                System.out.println("heartbeat: " + record.getEvent().getHeartbeatEvent().getEpoch());
                break;
            case CHECKPOINT:
                SubscribeProtobufData.CheckpointEvent evt = record.getEvent().getCheckpointEvent();
                System.out.println("Checkpoint: gtid=" + evt.getSyncedGtid()
                        + ", file=" + evt.getFileName() + ", position=" + evt.getPosition());
                break;
            case BEGIN:
                System.out.println("BEGIN");
                break;
            case COMMIT:
                System.out.println("COMMIT");
                break;
            case ROLLBACK:
                System.out.println("ROLLBACK");
                break;
            case DDL:
                System.out.println(record.getEvent().getDdlEvent().getSql());
                break;
            case DML:
                switch (record.getEvent().getDmlEvent().getDmlEventType()) {
                    case INSERT:
                        transInsert(record);
                        break;
                    case DELETE:
                        transDelete(record);
                        break;
                    case UPDATE:
                        transUpdate(record);
                        break;
                    default:
                        throw new IllegalStateException("unsupported dml event type");
                }
                break;
            default:
                throw new IllegalStateException("unsupported message type");
        }
    }

    private static String decode(SubscribeProtobufData.Data data) throws Exception {
        switch (data.getDataType()) {
            case INT8:
            case INT16:
            case INT32:
            case INT64:
            case UINT8:
            case UINT16:
            case UINT32:
            case UINT64:
            case DECIMAL:
            case FLOAT32:
            case FLOAT64:
                return data.getSv();
            case STRING:
                return "_binary'" + new String(data.getBv().toByteArray()) + "'";
            case BYTES:
                return "x'" + DatatypeConverter.printHexBinary(data.getBv().toByteArray()) + "'";
            case NA:
                return "DEFAULT";
            case NIL:
                return "NULL";
            default:
                throw new IllegalStateException("unsupported data type");
        }
    }

    private static void transUpdate(SubscribeProtobufData.Entry entry) throws Exception {
        SubscribeProtobufData.DMLEvent dmlEvt = entry.getEvent().getDmlEvent();
        String header = String.format(
                "UPDATE `%s`.`%s` SET ",
                entry.getHeader().getSchemaName(),
                entry.getHeader().getTableName());
        for (SubscribeProtobufData.RowChange row : dmlEvt.getRowsList()) {
            System.out.print(header);
            String sep = "";
            for (int i = 0; i < row.getNewColumnsCount(); i++) {
                SubscribeProtobufData.Data col = row.getNewColumns(i);
                if (col.getDataType() == SubscribeProtobufData.DataType.NA) {
                    continue;
                }

                System.out.print(sep);
                SubscribeProtobufData.Column colDef = dmlEvt.getColumns(i);
                System.out.print(String.format("`%s`", colDef.getName()));
                if (col.getDataType() == SubscribeProtobufData.DataType.NIL) {
                    System.out.print(" IS NULL");
                } else {
                    System.out.print(String.format(" = %s", decode(col)));
                }
                sep = ", ";
            }

            System.out.print(" WHERE ");
            sep = "";
            for (int i = 0; i < row.getOldColumnsCount(); i++) {
                SubscribeProtobufData.Data col = row.getOldColumns(i);
                if (col.getDataType() == SubscribeProtobufData.DataType.NA) {
                    continue;
                }
                System.out.print(sep);
                SubscribeProtobufData.Column colDef = dmlEvt.getColumns(i);
                System.out.print(String.format("`%s`", colDef.getName()));
                if (col.getDataType() == SubscribeProtobufData.DataType.NIL) {
                    System.out.print(" IS NULL");
                } else {
                    System.out.print(String.format(" = %s", decode(col)));
                }
                sep = " AND ";
            }
            System.out.println(" LIMIT 1");
        }
    }


    private static void transDelete(SubscribeProtobufData.Entry entry) throws Exception {
        SubscribeProtobufData.DMLEvent dmlEvt = entry.getEvent().getDmlEvent();
        String header = String.format(
                "DELETE FROM `%s`.`%s` WHERE ",
                entry.getHeader().getSchemaName(),
                entry.getHeader().getTableName());
        for (SubscribeProtobufData.RowChange row : dmlEvt.getRowsList()) {
            System.out.print(header);
            String sep = "";
            for (int i = 0; i < row.getOldColumnsCount(); i++) {
                SubscribeProtobufData.Data col = row.getOldColumns(i);
                if (col.getDataType() == SubscribeProtobufData.DataType.NA) {
                    continue;
                }

                System.out.print(sep);
                SubscribeProtobufData.Column colDef = dmlEvt.getColumns(i);
                System.out.print(String.format("`%s`", colDef.getName()));
                if (col.getDataType() == SubscribeProtobufData.DataType.NIL) {
                    System.out.print(" IS NULL");
                } else {
                    System.out.print(String.format(" = %s", decode(col)));
                }
                sep = " AND ";
            }
            System.out.println(" LIMIT 1");
        }
    }

    private static void transInsert(SubscribeProtobufData.Entry entry) throws Exception {
        SubscribeProtobufData.DMLEvent dmlEvt = entry.getEvent().getDmlEvent();
        System.out.print(
                String.format(
                        "INSERT INTO `%s`.`%s` VALUES ",
                        entry.getHeader().getSchemaName(),
                        entry.getHeader().getTableName()));
        String rowSep = "";
        for (SubscribeProtobufData.RowChange row : dmlEvt.getRowsList()) {
            System.out.print(rowSep);
            System.out.print('(');
            String sep = "";
            for (SubscribeProtobufData.Data col : row.getNewColumnsList()) {
                System.out.print(sep);
                System.out.print(decode(col));
                sep = ", ";
            }
            System.out.print(')');
            rowSep = ", ";
        }

        System.out.println();
    }
}
